; SPDX-License-Identifier: MIT
;
;	@file		ntv2vcam.def
;	@brief		Implementation of NTV2VCAM class.
;	@copyright	(C) 2025 AJA Video Systems, Inc.  All rights reserved.
;

LIBRARY ntv2vcam
EXPORTS
    DllMain             PRIVATE
    DllGetClassObject   PRIVATE
    DllCanUnloadNow     PRIVATE
    DllRegisterServer   PRIVATE
    DllUnregisterServer PRIVATE
