/* SPDX-License-Identifier: MIT */
/**
	@file		public.h
	@brief		Master header for the ajabase library.
	@copyright	(C) 2009-2022 AJA Video Systems, Inc.  All rights reserved.
**/

#ifndef AJA_PUBLIC_H
	#define AJA_PUBLIC_H

	#include "ajabase/common/export.h"
	#include <string>
	#include <list>
	#include <vector>
	#include <map>
	#include "ajabase/common/types.h"
	#if defined(AJA_USE_CPLUSPLUS11)
		#include <mutex>
	#endif

#endif	//	AJA_PUBLIC_H
