/*
 * SPDX-License-Identifier: MIT
 * Copyright (C) 2004 - 2022 AJA Video Systems, Inc.
 */
///////////////////////////////////////////////////////////////
//
// NTV2 Linux v2.6 Device Driver for AJA boards.
//
////////////////////////////////////////////////////////////
//
// Filename: ntv2driverbigphysarea.h
// Purpose:	 Header file to enable support for optional BIGPHYSAREA kernel patch.
//
///////////////////////////////////////////////////////////////

#ifndef NTV2DRIVERBIGPHYSAREA_HEADER
#define NTV2DRIVERBIGPHYSAREA_HEADER

// Define this only if you've installed the bigphysarea patch
// If not, comment out only this single line: 
//#define USE_BIGPHYSAREA

#endif

