#define IDD_MYFILTER_PROPPAGE 100

#define IDC_STATIC_AJADEVICE 200
#define IDC_STATIC_INPUTYPE 201
#define IDC_STATIC_INPUTCHANNEL 202
#define IDC_STATIC_PIXELFORMAT 203

#define IDC_EDIT_AJADEVICE 300
#define IDC_EDIT_INPUTTYPE 301
#define IDC_EDIT_INPUTCHANNEL 302
#define IDC_EDIT_PIXELFORMAT 303

#define IDS_MYFILTER_PROPNAME 500
