/* SPDX-License-Identifier: MIT */
/*
  This software is provided by AJA Video, Inc. "AS IS"
  with no express or implied warranties.
*/

#ifndef _NTV2_GL_TEXTURE_TRANSFER_NV_
#define _NTV2_GL_TEXTURE_TRANSFER_NV_


#include "ntv2glTextureTransfer.h"


CNTV2glTextureTransfer *CreateNTV2glTextureTransferNV();

#endif

