#include <windows.h>
#include "resource.h"

IDD_MYFILTER_PROPPAGE DIALOGEX 0, 0, 239, 124
STYLE WS_CHILD
FONT 8, "MS Sans Serif", 0, 0, 0x0
BEGIN
    LTEXT           "AJA device",IDC_STATIC_AJADEVICE,0,9,80,8
    EDITTEXT        IDC_EDIT_AJADEVICE,90,9,148,14,ES_AUTOHSCROLL
    LTEXT           "Input Type",IDC_STATIC_INPUTYPE,0,31,80,8
    EDITTEXT        IDC_EDIT_INPUTTYPE,90,31,148,14,ES_AUTOHSCROLL
    LTEXT           "Input Channel",IDC_STATIC_INPUTCHANNEL,0,54,80,8
    EDITTEXT        IDC_EDIT_INPUTCHANNEL,90,54,148,14,ES_AUTOHSCROLL
    LTEXT           "Pixel Format",IDC_STATIC_PIXELFORMAT,0,79,80,8
    EDITTEXT        IDC_EDIT_PIXELFORMAT,90,79,148,14,ES_AUTOHSCROLL
    DEFPUSHBUTTON   "Apply", IDOK, 189, 110, 50, 14
END

STRINGTABLE DISCARDABLE
BEGIN
    IDS_MYFILTER_PROPNAME "Config"
END