/* SPDX-License-Identifier: MIT */
/**
	@file		cli_utils.h
	@brief		Declaration of GetNumber, GetString and GetCharAsInt functions.
	@copyright	(C) 2011-2022 AJA Video Systems, Inc.  All rights reserved.
**/

#ifndef AJA_CLI_UTILS_H
#define AJA_CLI_UTILS_H

/*
 * This file has been removed the functionality is available elsewhere in ajabase
 * see common.h
 */

#endif // AJA_CLI_UTILS_H
