/*
 * SPDX-License-Identifier: MIT
 * Copyright (C) 2004 - 2022 AJA Video Systems, Inc.
 */
///////////////////////////////////////////////////////////////
//
// HEVC Linux Device Driver for AJA HEVC boards.
//
////////////////////////////////////////////////////////////
//
// Filename: hevccommon.h
// Purpose:	 Common internal header file for driver.
// Notes:	 
//
///////////////////////////////////////////////////////////////

#ifndef HEVCCOMMON_H
#define HEVCCOMMON_H

#include "../ntv2system.h"
#include "hevcpublic.h"
#include "hevcconstants.h"
#include "hevcparams.h"

#endif
