/* SPDX-License-Identifier: MIT */
/**
	@file		ntv2enumerateboards.h
	@brief		Header file that defines the NTV2EnumerateDevices class
	@copyright	(C) 2013-2022 AJA Video Systems, Inc.  All rights reserved.
**/
